﻿using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.BioBoost.ProblemDescription;
using HeuristicLab.BioBoost.Representation;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.BioBoost.Operators.Mutation {
  [StorableClass]
  public abstract class PlantBasedUtilizationModifier : SingleSuccessorOperator, IManipulator, IStochasticOperator {

    #region Parameters
    public LookupParameter<BioBoostProblemData> ProblemDataParameter { get { return (LookupParameter<BioBoostProblemData>) Parameters["ProblemData"]; } }
    public LookupParameter<DistanceMatrix> DistanceMatrixParameter { get { return (LookupParameter<DistanceMatrix>) Parameters["DistanceMatrix"]; } }
    public ValueParameter<BoolValue> DistanceMatrixInProblemDataParameter { get { return (ValueParameter<BoolValue>) Parameters["DistanceMatrixInProblemData"]; } }
    public ValueLookupParameter<IntMatrix> RegionBoundsParameter { get { return (ValueLookupParameter<IntMatrix>) Parameters["RegionBounds"]; } }
    public ILookupParameter<IRandom> RandomParameter { get { return (ILookupParameter<IRandom>) Parameters["Random"]; } }
    public IValueLookupParameter<DoubleMatrix> BoundsParameter { get { return (IValueLookupParameter<DoubleMatrix>) Parameters["Bounds"]; } }
    #endregion

    #region Parameter Values
    public BioBoostProblemData ProblemData { get { return ProblemDataParameter.ActualValue; } }
    public IntMatrix RegionBounds { get { return RegionBoundsParameter.ActualValue; } }
    public DoubleMatrix Bounds { get { return BoundsParameter.ActualValue; } }
    public IRandom Random { get { return RandomParameter.ActualValue; } }
    public bool DistanceMatrixInProblemData { get { return DistanceMatrixInProblemDataParameter.Value.Value; } }
    public DistanceMatrix DistanceMatrix {
      get {
        if (DistanceMatrixInProblemData)
          return ((IValueParameter<DistanceMatrix>) ProblemData.Parameters[DistanceMatrixParameter.ActualName]).Value;
        else return DistanceMatrixParameter.ActualValue;
      }
    }
    #endregion

    #region Construction & Cloning

    [StorableConstructor]
    public PlantBasedUtilizationModifier(bool isDeserializing) {}
    public PlantBasedUtilizationModifier(PlantBasedUtilizationModifier orig, Cloner cloner) : base(orig, cloner) {}
    public PlantBasedUtilizationModifier() {
      Parameters.Add(new LookupParameter<BioBoostProblemData>("ProblemData", "The data store of the detailed problem description."));
      Parameters.Add(new LookupParameter<DistanceMatrix>("DistanceMatrix", "The distance matrix to use", "StreetDistanceMatrix")); // TODO: check this
      Parameters.Add(new ValueParameter<BoolValue>("DistanceMatrixInProblemData", "Whether to look for the distance matrix in problem data or in scope.", new BoolValue(true)));
      Parameters.Add(new ValueLookupParameter<IntMatrix>("RegionBounds", "The limits of valid region ids."));
      Parameters.Add(new ValueLookupParameter<DoubleMatrix>("Bounds", "The limits for manipulating utilizations.", new DoubleMatrix(new [,]{{0d, 1d}})));
      Parameters.Add(new LookupParameter<IRandom>("Random", "The random number generator."));
    }

    #endregion

    public override IOperation Apply() {
      var scope = ExecutionContext.Scope;
      var productLinks = ProblemData.ProductLinks;
      var feedstock =
        ProblemData.TransportTargets.CheckedItems.ElementAt(
          Random.Next(ProblemData.TransportTargets.CheckedItems.Count())).Value.Value;
      var supplyTransports = scope.Variables[LayerDescriptor.TransportTargets.NameWithPrefix(feedstock)].Value as IntegerVector;
      if (supplyTransports == null) return base.Apply();
      var plants = supplyTransports.Select((target, source) => new {target, source}).GroupBy(p => p.target).ToList();
      var plant = plants.ElementAt(Random.Next(plants.Count));
      var suppliers = plant.Select(s => s.source).ToList();
      var utilizations = GetUtilizations(scope, feedstock);
      var preceedingFeedstock = GetPreceedingFeedstock(scope, feedstock, productLinks);
      while (utilizations == null && preceedingFeedstock != null) {
        var transportTargets = GetTransportTargets(scope, preceedingFeedstock);
        if (transportTargets == null)
          throw new InvalidOperationException("Error preceeding feedstock \"" + preceedingFeedstock + "\" of \"" + feedstock + "\" has no transport targets");
        suppliers = // collect supplier's suppliers
          suppliers.Select(s => transportTargets.Select((target, source) => new {source, target}).Where(p => p.target == s).Select(p => p.source))
            .Aggregate((x, y) => x.Concat(y))
            .ToList();
        utilizations = GetUtilizations(scope, preceedingFeedstock);
        preceedingFeedstock = GetPreceedingFeedstock(scope, preceedingFeedstock, productLinks);
      }
      if (utilizations == null) {
        throw new InvalidOperationException("Error in supply chain: product " + feedstock +
                                            " does not have an underlying feedstock");
      }
      if (suppliers.Count > 0)
        Redistribute(utilizations, suppliers);
      return base.Apply();
    }

    private string GetPreceedingFeedstock(IScope scope, string feedstock, StringMatrix productLinks) {
      for (int i = 0; i < productLinks.Rows; i++) {
        if (productLinks[i, 1] == feedstock)
          return productLinks[i, 0];
      }
      return null;
    }

    private IntegerVector GetTransportTargets(IScope scope, string feedstock) {
      IVariable v = null;
      if (scope.Variables.TryGetValue(LayerDescriptor.TransportTargets.NameWithPrefix(feedstock), out v))
        return v.Value as IntegerVector;
      return null;
    }


    private RealVector GetUtilizations(IScope scope, string feedstock) {
      IVariable v = null;
      if (scope.Variables.TryGetValue(LayerDescriptor.Utilizations.NameWithPrefix(feedstock), out v))
        return v.Value as RealVector;
      return null;
    }

    public abstract void Redistribute(RealVector utilizations, List<int> supplierIndices);

    public double EnforceBounds(double value, int idx) {
      var min = Bounds[idx%Bounds.Rows, 0];
      var max = Bounds[idx%Bounds.Rows, 1];
      if (value < min) return min;
      if (value > max) return max;
      return value;
    }
  }
}